# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING

import responses

if TYPE_CHECKING:
    from ansys.simai.core.data.workspaces import Workspace


@responses.activate
def test_workspace_download_mer_data(simai_client):
    """WHEN downloading mer csv file
    THEN the content of the file matches the content of the response.
    """

    workspace: Workspace = simai_client._workspace_directory._model_from(
        {"id": "0011", "name": "riri"}
    )
    responses.add(
        responses.GET,
        f"https://test.test/workspaces/{workspace.id}/mer-data",
        body=b"mer-data-geometries",
        status=200,
    )

    in_memory = workspace.download_mer_data()
    data_in_file = in_memory.readline()
    assert data_in_file.decode("ascii") == "mer-data-geometries"



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import contextlib
import os
import ssl
import subprocess
import sys
import threading
from http.server import HTTPServer, SimpleHTTPRequestHandler
from unittest.mock import patch

import pytest

import ansys.simai.core.errors as err
from ansys.simai.core import SimAIClient


#
# SETUP
#
@pytest.fixture(scope="module")
def tls_root_certificate(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("tls_root_certificate")
    # Create OpenSSL config file for the CA with key usage extension
    with open(tmp_path / "ca_openssl.cnf", "w") as f:
        f.write("""
[req]
distinguished_name = req_distinguished_name
req_extensions = v3_ca
x509_extensions = v3_ca

[req_distinguished_name]

[v3_ca]
subjectKeyIdentifier = hash
authorityKeyIdentifier = keyid:always,issuer:always
basicConstraints = critical, CA:true
keyUsage = critical, digitalSignature, cRLSign, keyCertSign
""")
    # Create CA key and certificate with extensions
    subprocess.run(["openssl", "req", "-x509", "-newkey", "rsa:2048", "-nodes", "-keyout", tmp_path / "ca_key.pem", "-out", tmp_path / "test_ca.pem", "-days", "3650", "-subj", "/CN=Test CA", "-config", tmp_path / "ca_openssl.cnf"], check=False)  # fmt: skip # noqa: S607, S603
    # Create server key and certificate
    subprocess.run(["openssl", "req", "-newkey", "rsa:2048", "-nodes", "-keyout", tmp_path / "server_key.pem", "-out", tmp_path / "server_csr.pem", "-subj", "/CN=localhost"], check=False)  # fmt: skip # noqa: S607, S603
    with open(tmp_path / "server_openssl.cnf", "w") as f:
        # Create a config file for OpenSSL to include the SAN and key usage extensions
        f.write("""
[req]
distinguished_name = req_distinguished_name

[req_distinguished_name]

[server_ext]
basicConstraints = CA:FALSE
keyUsage = critical, digitalSignature, keyEncipherment
extendedKeyUsage = serverAuth
subjectAltName = DNS:localhost
""")
    subprocess.run(["openssl", "x509", "-req", "-in", tmp_path / "server_csr.pem", "-CA", tmp_path / "test_ca.pem", "-CAkey", tmp_path / "ca_key.pem", "-CAcreateserial", "-out", tmp_path / "server_cert.pem", "-days", "3650", "-extensions", "server_ext", "-extfile", tmp_path / "server_openssl.cnf"], check=False)  # fmt: skip # noqa: S607, S603

    return {
        "ca": tmp_path / "test_ca.pem",
        "server_key": tmp_path / "server_key.pem",
        "server_cert": tmp_path / "server_cert.pem",
    }


@pytest.fixture(scope="module")
def https_server(tls_root_certificate):
    class SimpleHTTPSHandler(SimpleHTTPRequestHandler):
        def do_GET(self):
            self.send_response(200)
            self.send_header("Content-type", "text/plain")
            self.end_headers()
            self.wfile.write(b"Hello, Secure World!")

    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_context.load_cert_chain(
        certfile=tls_root_certificate["server_cert"], keyfile=tls_root_certificate["server_key"]
    )
    httpd = HTTPServer(("localhost", 48219), SimpleHTTPSHandler)
    httpd.socket = ssl_context.wrap_socket(
        httpd.socket,
        server_side=True,
    )
    server_thread = threading.Thread(target=httpd.serve_forever, daemon=True)
    server_thread.start()
    yield "https://localhost:48219"
    httpd.shutdown()


BASE_CLT_ARGS = {
    "url": "https://test.example.com",
    "organization": "dummy",
    "_disable_authentication": True,
    "no_sse_connection": True,
    "skip_version_check": True,
}


def disable_http_retry(clt: SimAIClient, url: str):
    "Avoids slow tests due to retry backoff"
    adapter = clt._api._session.get_adapter(url)
    adapter.max_retries = 0


#
# TESTS
#
def test_client_without_config_tls_ca_bundle(tls_root_certificate, https_server):
    # Default config, using certifi's certificate store
    clt = SimAIClient(**BASE_CLT_ARGS)
    disable_http_retry(clt, https_server)
    # By default, the self-signed test CA is not accepted
    with pytest.raises(err.ConnectionError):
        clt._api._get(https_server)
    # One can use REQUESTS_CA_BUNDLE
    with patch.dict(os.environ, {"REQUESTS_CA_BUNDLE": str(tls_root_certificate["ca"])}):
        clt._api._get(https_server, return_json=False)


@pytest.mark.skipif(
    sys.version_info < (3, 11),
    reason='"system" requires Python >= 3.10, "patch" is broken in python 3.10',
)
def test_client_config_tls_ca_bundle_system(tls_root_certificate, https_server):
    clt = SimAIClient(**BASE_CLT_ARGS, tls_ca_bundle="system")
    disable_http_retry(clt, https_server)
    # The system CA rejects the test CA by default
    with pytest.raises(err.ConnectionError):
        clt._api._get(https_server)

    # If the host system trusts the test CA, pysimai trusts it !
    @contextlib.contextmanager
    def load_test_ca_as_truststore_system_ca(ctx: "ssl.SSLContext"):
        ctx.load_verify_locations(cafile=tls_root_certificate["ca"])
        yield

    with patch(
        "truststore._api._configure_context", side_effect=load_test_ca_as_truststore_system_ca
    ):
        clt._api._get(https_server, return_json=False)


@pytest.mark.skipif(
    sys.version_info >= (3, 10), reason="The error is only raised for python < 3.10"
)
def test_client_config_tls_ca_bundle_system_on_usupported_python_version():
    with pytest.raises(err.ConfigurationError, match="python >= 3.10"):
        SimAIClient(**BASE_CLT_ARGS, tls_ca_bundle="system")


def test_client_config_tls_ca_bundle_unsecure_none(https_server):
    clt = SimAIClient(**BASE_CLT_ARGS, tls_ca_bundle="unsecure-none")
    clt._api._get(https_server, return_json=False)


def test_client_config_tls_ca_bundle_path(tls_root_certificate, https_server):
    clt = SimAIClient(**BASE_CLT_ARGS, tls_ca_bundle=tls_root_certificate["ca"])
    clt._api._get(https_server, return_json=False)



"""
.. _ref_list_based_subset_assignment:

List-based subset assignment
============================

This example demonstrates how to distribute your dataset between Test
and Training subsets using lists.

"""

###############################################################################
# Import necessary libraries
# --------------------------

import ansys.simai.core

###############################################################################
# Create lists
# ------------
# List the data to be used for the test set.

# CAUTION:
# All training data that are not included into the following list will be
# assigned the training subset
TEST_LIST = ["My_td_1", "My_td_2"]

###############################################################################
# Connect to the platform
# ------------------------
# Connect to the SimAI platform. Refer to the :ref:`anchor-credentials`
# section of the documentation to adapt the connection type.

simai = ansys.simai.core.SimAIClient(organization="My_organization_name")
project = simai.projects.get(name="My_project_name")

###############################################################################
# Assign subsets
# --------------
# Assign a subset to each dataset (list) you created.

td_list = project.data
for td in td_list:
    if td.name in TEST_LIST:
        td.assign_subset(project, "Test")
    else:
        td.assign_subset(project, "Training")



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ruff: noqa: S311

import random
import threading
import time
from collections.abc import Callable
from sys import modules

import niquests
import pytest
import requests
from niquests.packages import urllib3

from ansys.simai.core import SimAIClient
from ansys.simai.core.api.client import ApiClient
from ansys.simai.core.data.geometries import Geometry, GeometryDirectory
from ansys.simai.core.data.models import Model
from ansys.simai.core.data.post_processings import (
    PostProcessing,
    PostProcessingDirectory,
)
from ansys.simai.core.data.predictions import Prediction
from ansys.simai.core.data.projects import Project
from ansys.simai.core.data.training_data import TrainingData
from ansys.simai.core.data.workspaces import Workspace
from ansys.simai.core.utils.configuration import ClientConfig

# HACK: make responses work with niquests, coming from the niquests docs
# https://niquests.readthedocs.io/en/stable/dev/migrate.html
modules["requests"] = niquests
modules["requests.adapters"] = niquests.adapters
modules["requests.exceptions"] = niquests.exceptions
modules["requests.compat"] = requests.compat
modules["requests.packages.urllib3"] = urllib3
###


@pytest.fixture(scope="session")
def api_client():
    """A ApiClient object with bogus URL"""
    yield ApiClient(
        config=ClientConfig(
            url="https://test.test",
            _disable_authentication=True,
            no_sse_connection=True,
            organization="ExtraClaquette",
        )
    )


# warning: even with scope=function, the same SimAIClient is kept,
# thus previous test data exist in different methods
@pytest.fixture(scope="function")
def simai_client():
    client = SimAIClient(
        url="https://test.test",
        _disable_authentication=True,
        no_sse_connection=True,
        skip_version_check=True,
        organization="ExtraCorp",
    )
    client._current_workspace = Workspace(
        client,
        client._workspace_directory,
        {"id": "whatever", "name": "not-an-actual-workspace-please-set-it-if-needed"},
    )
    yield client


@pytest.fixture(scope="function")
def sse_mixin(simai_client):
    yield ApiClient(
        simai_client=simai_client,
        config=ClientConfig(
            url="https://test.test",
            _disable_authentication=True,
            no_sse_connection=True,
            organization="PhantomOrg",
        ),
    )


@pytest.fixture(scope="function")
def post_processing_factory(simai_client) -> PostProcessing:
    """Returns a function to create a :py:class:`PostProcessing`."""

    def _factory(prediction=None, **kwargs) -> PostProcessing:
        kwargs.setdefault("id", str(random.random()))
        if prediction and "prediction_id" not in kwargs:
            kwargs["prediction_id"] = prediction.id
        kwargs.setdefault("state", "successful")
        return simai_client._post_processing_directory._model_from(kwargs, prediction=prediction)

    return _factory


@pytest.fixture(scope="function")
def prediction_factory(simai_client) -> Prediction:
    """Returns a function to create a :py:class:`Prediction`."""

    def _factory(post_processings=None, geometry=None, **kwargs) -> Prediction:
        if "id" not in kwargs:
            if "boundary_conditions" in kwargs:
                kwargs["id"] = "pred-" + "-".join(
                    str(s) for s in kwargs["boundary_conditions"].values()
                )
            else:
                kwargs["id"] = str(random.random())
        if geometry and "geometry_id" not in kwargs:
            kwargs["geometry_id"] = geometry.id
        kwargs.setdefault("boundary_conditions", {"Vx": 10.01, "Vy": 0.0009})
        kwargs.setdefault("state", "successful")
        kwargs.setdefault("confidence_score", None)
        kwargs.setdefault("raw_confidence_score", None)
        prediction = simai_client._prediction_directory._model_from(kwargs)
        if post_processings is not None:
            # If we passed post-processings as parameter,
            # link them with the prediction.
            # Evidently PP should not be passed directly if the aim
            # is to test creation requests.
            prediction.get_post_processings = lambda: post_processings
            for pp in post_processings:
                pp._prediction = prediction

                # add post processing to prediction
                pred_pp = prediction._post_processings
                params = pp.fields.get("location", {})
                params_frozen = frozenset(params.items())
                pp_type = PostProcessingDirectory._data_model_for_type_name(pp._api_name())
                if pp_type not in pred_pp._post_processings:
                    pred_pp._post_processings[pp_type] = {}
                pred_pp._post_processings[pp_type][params_frozen] = pp
        return prediction

    return _factory


@pytest.fixture()
def geometry_factory(simai_client) -> Geometry:
    """Returns a function to create a :py:class:`Geometry`."""

    def _factory(predictions=None, **kwargs) -> Geometry:
        kwargs.setdefault("id", str(random.random()))
        kwargs.setdefault("name", kwargs["id"])
        kwargs.setdefault("state", "successful")
        if not predictions:
            predictions = []
        kwargs["predictions"] = [s.id for s in predictions]

        geometry = simai_client._geometry_directory._model_from(kwargs)
        geometry.get_predictions = lambda: predictions
        return geometry

    return _factory


@pytest.fixture()
def training_data_factory(simai_client) -> TrainingData:
    """Returns a function to create a :py:class:`TrainingData`."""

    def _factory(**kwargs) -> TrainingData:
        kwargs.setdefault("id", str(random.random()))
        kwargs.setdefault("name", kwargs["id"])
        kwargs.setdefault("state", "successful")

        training_data = simai_client._training_data_directory._model_from(kwargs)
        return training_data

    return _factory


@pytest.fixture()
def project_factory(simai_client) -> Project:
    """Returns a function to create a :py:class:`Project`."""

    def _factory(**kwargs) -> Project:
        kwargs.setdefault("id", str(random.random()))
        kwargs.setdefault("name", kwargs["id"])

        project = simai_client._project_directory._model_from(kwargs)
        return project

    return _factory


@pytest.fixture()
def geometry_directory(simai_client):
    yield GeometryDirectory(client=simai_client)


@pytest.fixture()
def create_mock_geometry():
    """Creates a lightweight geometry, not tracked by simai_client"""

    def _factory(id, predictions=None, **kwargs):
        geometry = Geometry(None, None, {"id": id, "name": str(id), "metadata": kwargs})
        geometry.get_predictions = lambda: predictions or []
        return geometry

    return _factory


@pytest.fixture(scope="function")
def global_coefficient_request_factory(simai_client):
    """Returns a function to create a global coefficient request object."""

    def _factory(**kwargs):
        return simai_client._process_gc_formula_directory._model_from(**kwargs)

    return _factory


@pytest.fixture()
def model_factory(simai_client) -> Model:
    """Returns a function to create a :py:class:`Model`."""

    def _factory(**kwargs) -> Model:
        kwargs.setdefault("id", str(random.random()))
        kwargs.setdefault("name", kwargs["id"])

        model = simai_client._model_directory._model_from(kwargs)
        return model

    return _factory


@pytest.fixture
def delayed_events():
    """
    Fixture to test Event.wait() with sequential events.

    This fixture provides a manager for scheduling events to happen at specific times,
    allowing for controlled testing of wait() behavior.

    Usage:
        def test_something(delayed_events):
            # Schedule events to occur
            delayed_events.add(lambda: model.update_state("processing"))
            delayed_events.add(lambda: model.complete())

            # Start the background thread that will trigger events
            delayed_events.start()

            # Call the wait method you're testing
            result = model.wait()

            # Make assertions
            assert result is True
    """

    class DelayedEventManager:
        def __init__(self) -> None:
            self.events = []
            self.thread = None

        def add(self, callback: Callable[[], None], delay: float = 0.1):
            """Add a callback to be executed after the specified delay (in seconds)"""
            self.events.append((delay, callback))

        def start(self):
            """Start the thread that will execute all scheduled events"""

            def run_events():
                start_time = time.time()

                for delay, callback in self.events:
                    time_to_wait = delay - (time.time() - start_time)
                    if time_to_wait > 0:
                        time.sleep(time_to_wait)
                    callback()

            self.thread = threading.Thread(target=run_events)
            self.thread.daemon = True
            self.thread.start()
            return self

        def join(self):
            """Wait for all events to complete"""
            if self.thread:
                self.thread.join()

    manager = DelayedEventManager()

    yield manager

    manager.join()



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from io import BytesIO

import pytest
import responses
from pydantic import HttpUrl
from responses import matchers

from ansys.simai.core import SimAIClient
from ansys.simai.core.errors import ApiClientError, NotFoundError


def test_construct_default_url():
    """WHEN No URL is passed to SimAIClient
    THEN A default one is used
    """
    client = SimAIClient(
        no_sse_connection=True,
        _disable_authentication=True,
        skip_version_check=True,
        organization="ExtraBanane",
    )
    assert client._api._url_prefix == HttpUrl("https://api.simai.ansys.com/v2/")


@responses.activate(registry=responses.registries.OrderedRegistry)
def test_retry_on_5XX(api_client):
    url = "https://try.me/"
    rsp1 = responses.add(responses.GET, url=url, status=503)
    rsp2 = responses.add(responses.GET, url=url, status=504)
    rsp3 = responses.add(responses.GET, url=url, status=200, json={"wat": "hyperdrama"})
    resp = api_client._get(url)
    assert resp == {"wat": "hyperdrama"}
    assert rsp1.call_count == 1
    assert rsp2.call_count == 1
    assert rsp3.call_count == 1


@responses.activate
def test_404_response_raises_not_found_error(api_client):
    """WHEN ApiClient gets a 404 response
    THEN a NotFoundError is raised
    """
    responses.add(
        responses.GET,
        "https://test.test/expected_format",
        json={"message": "beep-boop"},
        status=404,
    )
    responses.add(
        responses.GET,
        "https://test.test/only_status_format",
        json={"status": "not found"},
        status=404,
    )
    responses.add(responses.GET, "https://test.test/no_content", status=404)

    with pytest.raises(NotFoundError) as exc_info:
        api_client._get("expected_format")
    assert exc_info.value.args[0] == "beep-boop"
    with pytest.raises(NotFoundError) as exc_info:
        api_client._get("only_status_format")
    assert exc_info.value.args[0] == "not found"
    with pytest.raises(NotFoundError) as exc_info:
        api_client._get("no_content")
    assert exc_info.value.args[0] == "Not Found"


@responses.activate
@pytest.mark.parametrize(
    "code, reason",
    [
        (400, "Bad Request"),
        (401, "Unauthorized"),
        (403, "Forbidden"),
        # (404, "Not Found"), # raises NotFoundError instead # noqa: ERA001
        (405, "Method Not Allowed"),
        (500, "Internal Server Error"),
        (501, "Not Implemented"),
        (502, "Bad Gateway"),
    ],
)
def test_4XX_5XX_responses_raise_api_client_error_no_json(api_client, code, reason):
    """WHEN ApiClient gets a 4XX or 5XX response without details in json
    THEN a ApiClientError is raised with the code and reason
    """
    responses.add(responses.GET, f"https://test.test/{code}", status=code)

    with pytest.raises(ApiClientError) as exc_info:
        api_client._get(f"{code}")
    assert exc_info.value.args[0] == f"{code} {reason}"


@responses.activate
@pytest.mark.parametrize(
    "json, message",
    [
        ({"errors": {"beep": "boop"}}, "{'beep': 'boop'}"),
        ({"message": "tomato"}, "tomato"),
        (
            {"status": "I'm sorry Dave, I'm afraid I can't do that"},
            "I'm sorry Dave, I'm afraid I can't do that",
        ),
        (None, "Bad Request"),
    ],
)
def test_4XX_5XX_responses_raise_api_client_error_with_json(api_client, json, message):
    """WHEN ApiClient gets a 4XX or 5XX response without details in json
    THEN a ApiClientError is raised with the code and message from the json or fallback
    """
    responses.add(responses.GET, "https://test.test/errors", json=json, status=400)

    with pytest.raises(ApiClientError) as exc_info:
        api_client._get("errors")
    assert exc_info.value.args[0] == f"400 {message}"


def test_use_system_proxies(mocker):
    """WHEN ApiClient is initialized on a system with default proxies
    THEN the system proxies are used
    """
    mocker.patch(
        "ansys.simai.core.api.mixin.getproxies",
        return_value={"grpc": "https://bonzibuddy.org"},
    )
    client = SimAIClient(
        no_sse_connection=True,
        _disable_authentication=True,
        skip_version_check=True,
        organization="ExtraBanane",
    )
    assert client._api._session.proxies == {"grpc": "https://bonzibuddy.org"}


def test_use_user_provided_proxies(mocker):
    """WHEN ApiClient is initialized with a specific https_proxy
    THEN the specific proxy is used (system proxies are ignored), the url is normalized
    """
    mocker.patch(
        "ansys.simai.core.api.mixin.getproxies",
        return_value={"grpc": "https://bonzibuddy.org"},
    )

    client = SimAIClient(
        no_sse_connection=True,
        _disable_authentication=True,
        skip_version_check=True,
        organization="ExtraBanane",
        https_proxy="https://😛.com",
    )
    assert client._api._session.proxies == {"https": "https://xn--528h.com/"}


@responses.activate
def test_upload_file_with_presigned_post_monitor_callback(mocker, api_client):
    presigned_post = {"fields": {"sandstorm": "TUTUTUUTUTU"}, "url": "https://leekspin.com/"}
    file = BytesIO(b"Hello World")
    file.name = "hello.txt"
    responses.add(
        responses.POST,
        "https://leekspin.com/",
        match=[
            matchers.multipart_matcher(
                data={
                    **presigned_post["fields"],
                },
                files={"file": ("hello.txt", file, "application/octet-stream")},
            )
        ],
        status=200,
    )
    monitor_values = []
    api_client.upload_file_with_presigned_post(
        file=file,
        presigned_post=presigned_post,
        monitor_callback=lambda x: monitor_values.append(x),
    )
    # Bigger than `file` because:
    # * the multipart encoding
    # * the other fields (`sandstorm`)
    assert monitor_values == [297]



"""
.. _ref_model_recomputation:

Model recomputation
===================

This example demonstrates how to relaunch a model build using the latest
model configuration in a same project.

"""

###############################################################################
# Import necessary libraries
# --------------------------

import ansys.simai.core as asc

simai = asc.from_config()

###############################################################################
# Get the project from the server
# -------------------------------

my_project = simai.projects.get(name="old-ps")

###############################################################################
# Get the model configuration
# ---------------------------
# Get the latest model configuration of the project.

last_build_config = my_project.last_model_configuration

###############################################################################
# Verify the project requirements
# -------------------------------
# Verify that the project meets the requirements for training (model building).

is_trainable_check = my_project.is_trainable()

###############################################################################
# If the project met the requirements, launch a model build.
# Otherwise, print the reasons the project does not meet the requirements.
if is_trainable_check:
    new_model = simai.models.build(last_build_config)
else:
    print(is_trainable_check.reason)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ruff: noqa: S105, S106

import copy
import json
import time
from datetime import datetime, timezone
from math import ceil

import niquests
import pytest
import responses
from responses.matchers import urlencoded_params_matcher

from ansys.simai.core.errors import SimAIError
from ansys.simai.core.utils.auth import Authenticator, _AuthTokens, _AuthTokensRetriever
from ansys.simai.core.utils.configuration import ClientConfig, Credentials

DEFAULT_TOKENS = {
    "access_token": "check",
    "expires_in": 60,
    "refresh_expires_in": 1800,
    "refresh_token": "kaboom",
}


@responses.activate
def test_request_auth_tokens_direct_grant_bad_credentials_raises(mocker, tmpdir):
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    responses.add(
        responses.POST,
        "http://myauthserver.com/protocol/openid-connect/token",
        json={
            "error": "invalid_grant",
            "error_description": "Invalid user credentials",
        },
        status=401,
        match=[
            urlencoded_params_matcher(
                {
                    "username": "macron",
                    "password": "explosion",
                    "client_id": "sdk",
                    "grant_type": "password",
                    "scope": "openid",
                }
            )
        ],
    )
    tokens_retriever = _AuthTokensRetriever(
        credentials=None,
        session=niquests.Session(),
        realm_url="http://myauthserver.com",
        auth_cache_hash="rando",
    )
    with pytest.raises(SimAIError, match="Invalid user credentials"):
        tokens_retriever.credentials = Credentials(username="macron", password="explosion")
        tokens_retriever.get_tokens()


@responses.activate
def test_request_auth_tokens_direct_grant(mocker, tmpdir):
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    responses.add(
        responses.POST,
        "http://myauthserver.com/protocol/openid-connect/token",
        json=DEFAULT_TOKENS,
        status=200,
        match=[
            urlencoded_params_matcher(
                {
                    "username": "timmy",
                    "client_id": "sdk",
                    "grant_type": "password",
                    "scope": "openid",
                }
            )
        ],
    )
    tokens_retriever = _AuthTokensRetriever(
        credentials=None,
        session=niquests.Session(),
        realm_url="http://myauthserver.com",
        auth_cache_hash="rando",
    )
    tokens_retriever.credentials = Credentials(username="timmy", password="")
    tokens = tokens_retriever.get_tokens()
    assert tokens.refresh_token == DEFAULT_TOKENS["refresh_token"]
    assert tokens.access_token == DEFAULT_TOKENS["access_token"]


@responses.activate
def test_token_refresh_failure_triggers_reauth(mocker, tmpdir):
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    resps_refresh = responses.add(
        responses.POST,
        "http://myauthserver.com/protocol/openid-connect/token",
        status=418,
        match=[
            urlencoded_params_matcher(
                {"client_id": "sdk", "grant_type": "refresh_token", "refresh_token": "revoked"}
            )
        ],
    )
    resps_direct_grant = responses.add(
        responses.POST,
        "http://myauthserver.com/protocol/openid-connect/token",
        json=DEFAULT_TOKENS,
        status=200,
        match=[
            urlencoded_params_matcher(
                {
                    "username": "timmy",
                    "client_id": "sdk",
                    "grant_type": "password",
                    "scope": "openid",
                }
            )
        ],
    )
    with open(tmpdir / "tokens-rando.json", "w") as f:
        f.write(
            _AuthTokens(
                access_token="",
                expires_in=0,
                refresh_expires_in=999,
                refresh_token="revoked",
            ).model_dump_json()
        )
    tokens_retriever = _AuthTokensRetriever(
        credentials=Credentials(username="timmy", password=""),
        session=niquests.Session(),
        realm_url="http://myauthserver.com",
        auth_cache_hash="rando",
    )
    tokens = tokens_retriever.get_tokens()
    assert tokens.refresh_token == DEFAULT_TOKENS["refresh_token"]
    assert tokens.access_token == DEFAULT_TOKENS["access_token"]
    assert resps_refresh.call_count == 1
    assert resps_direct_grant.call_count == 1


@responses.activate
def test_request_auth_tokens_device_grant_with_bad_cache(mocker, tmpdir):
    """WHEN A device auth flow is requested
    AND a (bad) refresh token is cached
    THEN SDK tries to use the cached refresh token
    AND requests a device auth flow if the token is invalid
    """
    webbrowser_open = mocker.patch("webbrowser.open")
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    mocker.patch("ansys.simai.core.utils.auth.DEVICE_AUTH_POLLING_INTERVAL", 0)
    refresh_token = "spaghetti"
    device_code = "foxtrot uniform charlie kilo"
    realm_url = "http://myauthserver.com/my-realm"
    token_retriever = _AuthTokensRetriever(
        credentials=None, session=niquests.Session(), realm_url=realm_url, auth_cache_hash="lol"
    )
    fake_cache = _AuthTokens(
        access_token="",
        expires_in=0,
        refresh_expires_in=999,
        refresh_token=refresh_token,
    )
    with open(tmpdir / "tokens-lol.json", "w") as f:
        json.dump(fake_cache.model_dump_json(), f)

    responses.add(  # error when app tries to use invalidated cached token
        responses.POST,
        token_retriever.token_url,
        json={"poop": "Ur token down the drain"},
        status=418,
        match=[
            urlencoded_params_matcher(
                {
                    "client_id": "sdk",
                    "grant_type": "refresh_token",
                    "refresh_token": refresh_token,
                }
            )
        ],
    )
    responses.add(  # device-auth flow start
        responses.POST,
        token_retriever.device_auth_url,
        json={
            "verification_uri": "nope",
            "user_code": "abcdefg",
            "verification_uri_complete": "nope?code=abcdefg",
            "device_code": device_code,
        },
        status=200,
        match=[urlencoded_params_matcher({"client_id": "sdk", "scope": "openid"})],
    )
    responses.add(  # device-auth flow end
        responses.POST,
        token_retriever.token_url,
        json=DEFAULT_TOKENS,
        status=200,
        match=[
            urlencoded_params_matcher(
                {
                    "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    "client_id": "sdk",
                    "device_code": device_code,
                }
            )
        ],
    )

    tokens = token_retriever.get_tokens()
    webbrowser_open.assert_called_with("nope?code=abcdefg")
    assert tokens.access_token == DEFAULT_TOKENS["access_token"]
    assert tokens.refresh_token == DEFAULT_TOKENS["refresh_token"]
    assert ceil(tokens.expires_in) == DEFAULT_TOKENS["expires_in"]
    assert ceil(tokens.refresh_expires_in) == DEFAULT_TOKENS["refresh_expires_in"]


@responses.activate
def test_refresh_auth_tokens(mocker, tmpdir):
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    base_token = copy.deepcopy(DEFAULT_TOKENS)
    base_token["refresh_token"] = "kabam"
    final_token = copy.deepcopy(base_token)
    final_token["access_token"] = "new tok"

    expired_token = _AuthTokens.model_validate(copy.deepcopy(base_token))
    expired_token.expiration = datetime(year=1970, month=1, day=1, tzinfo=timezone.utc)
    token_retriever = _AuthTokensRetriever(
        credentials=None,
        session=niquests.Session(),
        realm_url="http://myauthserver.com",
        auth_cache_hash="popo",
    )
    token_retriever._get_token_from_cache = lambda: expired_token

    responses.add(
        responses.POST,
        "http://myauthserver.com/protocol/openid-connect/token",
        json=final_token,
        status=200,
        match=[
            urlencoded_params_matcher(
                {
                    "client_id": "sdk",
                    "grant_type": "refresh_token",
                    "refresh_token": "kabam",
                }
            )
        ],
    )

    fetched_token = token_retriever.get_tokens()
    assert fetched_token.access_token == final_token["access_token"]
    assert fetched_token.refresh_token == final_token["refresh_token"]


@responses.activate
def test_authenticator_automatically_refreshes_auth_before_requests_if_needed(mocker, tmpdir):
    with responses.RequestsMock(assert_all_requests_are_fired=True) as rsps:
        # Authentication request
        auth_tokens = copy.deepcopy(DEFAULT_TOKENS)
        auth_tokens["expires_in"] = 1
        auth_tokens["refresh_expires_in"] = 1800
        rsps.add(
            responses.POST,
            "https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
            json=auth_tokens,
            status=200,
            match=[
                urlencoded_params_matcher(
                    {
                        "client_id": "sdk",
                        "grant_type": "password",
                        "scope": "openid",
                        "username": "timmy",
                        "password": "key",
                    }
                )
            ],
        )
        # Token refresh request
        refresh_tokens = copy.deepcopy(DEFAULT_TOKENS)
        refresh_tokens["access_token"] = "check 1 2"
        refresh_tokens["refresh_token"] = "megazaur"
        rsps.add(
            responses.POST,
            "https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
            json=refresh_tokens,
            status=200,
            match=[
                urlencoded_params_matcher(
                    {
                        "client_id": "sdk",
                        "grant_type": "refresh_token",
                        "refresh_token": "kaboom",
                    }
                )
            ],
        )

        mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
        mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
        auth = Authenticator(
            ClientConfig(
                url="https://simai.ansys.com",
                organization="13_monkeys",
                credentials=Credentials(username="timmy", password="key"),
            ),
            niquests.Session(),
        )

        request = niquests.Request("GET", "https://simai.ansys.com/v2/models", auth=auth).prepare()
        assert request.headers.get("Authorization") == "Bearer check 1 2"
        assert request.headers.get("X-Org") == "13_monkeys"


@responses.activate
def test_authenticator_automatically_refreshes_auth_before_refresh_token_expires(mocker, tmpdir):
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)

    auth_tokens = copy.deepcopy(DEFAULT_TOKENS)
    auth_tokens["access_token"] = "monkey-see"
    auth_tokens["expires_in"] = _AuthTokens.EXPIRATION_BUFFER - 1
    auth_tokens["refresh_expires_in"] = _AuthTokensRetriever.REFRESH_BUFFER + 1
    resps_direct_grant = responses.add(
        responses.POST,
        "https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
        json=auth_tokens,
        status=200,
        match=[
            urlencoded_params_matcher(
                {
                    "client_id": "sdk",
                    "grant_type": "password",
                    "scope": "openid",
                    "username": "timmy",
                    "password": "key",
                }
            )
        ],
    )
    # Token refresh request
    refreshed_tokens = DEFAULT_TOKENS.copy()
    refreshed_tokens["access_token"] = "TFou"
    resps_refresh = responses.add(
        responses.POST,
        "https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
        json=refreshed_tokens,
        status=200,
        match=[
            urlencoded_params_matcher(
                {
                    "client_id": "sdk",
                    "grant_type": "refresh_token",
                    "refresh_token": "kaboom",
                }
            )
        ],
    )
    Authenticator(
        ClientConfig(
            url="https://simai.ansys.com",
            organization="14_monkeys",
            credentials=Credentials(username="timmy", password="key"),
        ),
        niquests.Session(),
    )
    assert resps_direct_grant.call_count == 1
    assert resps_refresh.call_count == 0
    t0 = time.time()
    while time.time() - t0 < 2 and resps_refresh.call_count == 0:
        # wait for the daemon thread to do its thing, or for the 2sec timeout...
        time.sleep(0.1)
    # Tokens are automatically refreshed so refresh token doesn't expire'
    assert resps_direct_grant.call_count == 1
    assert resps_refresh.call_count == 1


@responses.activate
def test_requests_outside_user_api_are_not_authentified(mocker, tmpdir):
    mocker.patch("ansys.simai.core.utils.auth.get_cache_dir", return_value=tmpdir)
    with responses.RequestsMock(assert_all_requests_are_fired=True) as rsps:
        # Authentication request
        keycloak_response_json = DEFAULT_TOKENS.copy()
        rsps.add(
            responses.POST,
            "https://simai.ansys.com/auth/realms/simai/protocol/openid-connect/token",
            json=keycloak_response_json,
            status=200,
        )

        auth = Authenticator(
            ClientConfig(
                url="https://simai.ansys.com",
                organization="Justice",
                credentials={"username": "timmy", "password": "D.A.N.C.E"},
            ),
            niquests.Session(),
        )
        request = niquests.Request("GET", "https://amazonaws.com/bloc-party", auth=auth).prepare()
        assert request.headers.get("Authorization") is None
        assert request.headers.get("X-Org") is None



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import pytest

from ansys.simai.core.data.model_configuration import ModelConfiguration, SupportedBuildPresets
from ansys.simai.core.errors import InvalidArguments


def test_build_preset_error(simai_client):
    """WHEN build_preset gets a non-supported value
    THEN an InvalidArgument is raised."""

    raw_project = {
        "id": "xX007Xx",
        "name": "fifi",
    }

    project = simai_client._project_directory._model_from(raw_project)

    with pytest.raises(InvalidArguments) as excinfo:
        ModelConfiguration(
            project=project,
            build_preset="not_valid_value",
        )

        assert f"{list(SupportedBuildPresets)}" in excinfo.value


def test_model_input_not_none(simai_client):
    """WHEN ModelConfiguration.input gets a None value
    THEN an InvalidArgument is raised."""

    raw_project = {
        "id": "xX007Xx",
        "name": "fifi",
    }

    project = simai_client._project_directory._model_from(raw_project)

    bld_conf = ModelConfiguration(
        project=project,
    )

    with pytest.raises(InvalidArguments):
        bld_conf.input = None


def test_model_output_not_none(simai_client):
    """WHEN ModelConfiguration.input gets a None value
    THEN an InvalidArgument is raised."""

    raw_project = {
        "id": "xX007Xx",
        "name": "fifi",
    }

    project = simai_client._project_directory._model_from(raw_project)

    bld_conf = ModelConfiguration(
        project=project,
    )

    with pytest.raises(InvalidArguments):
        bld_conf.output = None



# Copyright (C) 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from typing import Any, BinaryIO, List, Optional, Tuple, Union

from pydantic import BaseModel, Field, PositiveInt, ValidationError

from ansys.simai.core.data.base import ComputableDataModel, Directory
from ansys.simai.core.data.geomai.workspaces import GeomAIWorkspace
from ansys.simai.core.data.types import (
    File,
    Identifiable,
    get_id_from_identifiable,
)
from ansys.simai.core.errors import InvalidArguments

logger = logging.getLogger(__name__)


class GeomAIPredictionConfiguration(BaseModel):
    """The configuration used to run a GeomAI prediction."""

    latent_params: List[float]
    """A list of floats that represent the position of the geometry in the latent space.

    These parameters describe the shape in a compressed form.
    The number of floats should match the ``nb_latent_param`` your model was requested with.

    Required.
    """
    resolution: Optional[Tuple[PositiveInt, PositiveInt, PositiveInt]] = None
    """A list of three integers defining the number of divisions along the X, Y, and Z axes.

    Higher values allow for more detailed geometry reconstruction, while lower values reduce detail but speed up computation.

    Defaults to ``[100, 100, 100]`` if none is provided.
    """
    margin: Optional[float] = Field(default=None, ge=0, le=1)
    """A float that sets the size of the isosurface for the reconstruction of the geometry.

    Note:
        A margin of 0.0 is highly recommended for non-expert users.

    If you are an expert user, and if the generated geometry is noisy,
    you can try to adjust both resolution and margin values to find
    the right balance between them to generate a smoother geometry.

    | A higher margin gives a coarser surface with less detail.
    | A lower margin produces a sharper surface.

    Defaults to 0 if none is provided.
    """

    def __init__(self, *args, **kwargs):
        """Raises :exc:`~ansys.simai.core.errors.InvalidArguments` if the input data cannot be validated to from a valid model."""
        try:
            super().__init__(*args, **kwargs)
        except ValidationError as e:
            raise InvalidArguments(e.errors(include_url=False)) from None


class GeomAIPrediction(ComputableDataModel):
    """Provides the local representation of a GeomAI prediction object."""

    @property
    def configuration(self) -> GeomAIPredictionConfiguration:
        """The configuration used to run the prediction."""
        return GeomAIPredictionConfiguration.model_construct(self.fields["configuration"])

    def delete(self) -> None:
        """Remove a prediction from the server."""
        self._client._api.delete_geomai_prediction(self.id)
        self._unregister()

    def download(self, file: Optional[File] = None) -> Union[None, BinaryIO]:
        """Download the file generated by the prediction.

        Args:
            file: Binary file-object or the path of the file to put the content into.

        Returns:
            ``None`` if a file is specified or a binary file-object otherwise.
        """
        return self._client._api.download_geomai_prediction(self.id, file)


class GeomAIPredictionDirectory(Directory[GeomAIPrediction]):
    """Provides a collection of methods related to GeomAI model predictions.

    This method is accessed through ``client.geomai.predictions``.

    Example:
        .. code-block:: python

            import ansys.simai.core

            simai = ansys.simai.core.from_config()
            simai.geomai.predictions.list()
    """

    _data_model = GeomAIPrediction

    def list(
        self, workspace: Optional[Identifiable[GeomAIWorkspace]] = None
    ) -> List[GeomAIPrediction]:
        """List all GeomAI predictions on the server that belong to the specified workspace or the configured one.

        Args:
            workspace: ID or :class:`model <.workspaces.GeomAIWorkspace>` of the workspace to list the predictions for.
                This parameter is necessary if no workspace is set for the client.
        """
        workspace_id = get_id_from_identifiable(
            workspace, default=self._client.geomai._current_workspace
        )
        raw_predictions = self._client._api.geomai_predictions(workspace_id)
        return list(map(self._model_from, raw_predictions))

    def get(self, id: str) -> GeomAIPrediction:
        """Get a specific GeomAI prediction object from the server by ID.

        Args:
            id: ID of the prediction.

        Returns:
            :class:`GeomAIPrediction` instance with the given ID if it exists.

        Raises:
            NotFoundError: No prediction with the given ID exists.
        """
        return self._model_from(self._client._api.get_geomai_prediction(id))

    def delete(self, prediction: Identifiable[GeomAIPrediction]) -> None:
        """Delete a specific GeomAI prediction from the server.

        Args:
            prediction: ID or :class:`model <GeomAIPrediction>` of the prediction.

        Raises:
            NotFoundError: No prediction with the given ID exists.
        """
        prediction_id = get_id_from_identifiable(prediction)
        self._client._api.delete_geomai_prediction(prediction_id)
        self._unregister_item_with_id(prediction_id)

    def run(  # noqa: D417
        self,
        configuration: Union[GeomAIPredictionConfiguration, dict[str, Any]],
        workspace: Optional[Identifiable[GeomAIWorkspace]] = None,
    ) -> GeomAIPrediction:
        """Run a prediction in the given workspace with the given configuration.

        Args:
            configuration: The configuration to run the prediction with.
            workspace: Optional ID or :class:`model <.workspaces.GeomAIWorkspace>` of the target workspace.
                Defaults to the current workspace if set.

        Returns:
            Created prediction object.

        Raises:
            ProcessingError: If the server failed to process the request.

        Examples:
            .. code-block:: python

                simai = ansys.simai.core.from_config()
                workspace = simai.geomai.workspaces.list()[0]
                prediction = simai.geomai.predictions.run(
                    dict(latent_params=[0.1, 1.2, 0.76], resolution=(100, 100, 100), margin=0.0),
                    workspace,
                )
        """
        if not isinstance(configuration, GeomAIPredictionConfiguration):
            try:
                configuration = GeomAIPredictionConfiguration.model_validate(configuration)
            except ValidationError as e:
                raise InvalidArguments(e.errors(include_url=False)) from None
        workspace_id = get_id_from_identifiable(
            workspace, default=self._client.geomai._current_workspace
        )
        return self._model_from(
            self._client._api.run_geomai_prediction(workspace_id, configuration.model_dump())
        )

    def download(
        self, prediction: Identifiable[GeomAIPrediction], file: Optional[File] = None
    ) -> Union[None, BinaryIO]:
        """Download the file generated by the prediction.

        Args:
            prediction: ID or :class:`model <GeomAIPrediction>` of the prediction.
            file: Binary file-object or the path of the file to put the content into.

        Returns:
            ``None`` if a file is specified or a binary file-object otherwise.
        """
        prediction_id = get_id_from_identifiable(prediction)
        return self._client._api.download_geomai_prediction(prediction_id, file)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ansys.simai.core.api.geomai.predictions import GeomAIPredictionClientMixin
from ansys.simai.core.api.geomai.projects import GeomAIProjectClientMixin
from ansys.simai.core.api.geomai.training_data import GeomAITrainingDataClientMixin
from ansys.simai.core.api.geomai.training_data_parts import GeomAITrainingDataPartClientMixin
from ansys.simai.core.api.geomai.workspaces import GeomAIWorkspaceClientMixin
from ansys.simai.core.api.geometry import GeometryClientMixin
from ansys.simai.core.api.optimization import OptimizationClientMixin
from ansys.simai.core.api.point_cloud import PointCloudClientMixin
from ansys.simai.core.api.post_processing import PostProcessingClientMixin
from ansys.simai.core.api.prediction import PredictionClientMixin
from ansys.simai.core.api.project import ProjectClientMixin
from ansys.simai.core.api.sse import SSEMixin
from ansys.simai.core.api.training_data import TrainingDataClientMixin
from ansys.simai.core.api.training_data_part import TrainingDataPartClientMixin
from ansys.simai.core.api.workspace import WorkspaceClientMixin


class ApiClient(
    GeometryClientMixin,
    PointCloudClientMixin,
    OptimizationClientMixin,
    PostProcessingClientMixin,
    ProjectClientMixin,
    PredictionClientMixin,
    SSEMixin,
    TrainingDataClientMixin,
    TrainingDataPartClientMixin,
    WorkspaceClientMixin,
    GeomAIPredictionClientMixin,
    GeomAIProjectClientMixin,
    GeomAITrainingDataClientMixin,
    GeomAITrainingDataPartClientMixin,
    GeomAIWorkspaceClientMixin,
):
    """Provides the low-level client that handles direct communication with the server."""



"""
.. _ref_subset_assignment:

Subset assignment
=================

This example demonstrates how to assign a subset
to a training data.

"""

###############################################################################
# Import necessary libraries
# --------------------------

import ansys.simai.core as asc

simai = asc.from_config()

###############################################################################
# Select a training data
# ----------------------
# Example of a training_data_id associated with a project_id.

training_data_id = "k4z77qzq"
project_id = "k9756vw0"

###############################################################################
# Get subset assignment
# ---------------------
# Get and print the current subset assigned for this training_data_id.

current_subset = simai.training_data.get(id=training_data_id).get_subset(project=project_id)
print(current_subset)

###############################################################################
# Assign a new subset (two options)
# ---------------------------------
# Manually assign a new subset to the training data.

simai.training_data.get(id=training_data_id).assign_subset(project=project_id, subset="Test")

###############################################################################
# Alternatively, use SubsetEnum to assign a valid enum value to the training data.

from ansys.simai.core.data.types import SubsetEnum

simai.training_data.get(id=training_data_id).assign_subset(
    project=project_id, subset=SubsetEnum.TEST
)
